class SiteContact():
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.firstName = kwargs.get('firstName')
        self.surname = kwargs.get('surname')
        self.fullname = kwargs.get('fullname')


from komora_syncer.config import get_config
import requests
import traceback

from komora_syncer.models.komora.organization import Organization
from komora_syncer.models.komora.region import Region
from komora_syncer.models.komora.district import District
from komora_syncer.models.komora.municipality import Municipality
from komora_syncer.models.komora.site import Site
from komora_syncer.models.komora.device import Device
from komora_syncer.models.komora.contact import Contact

from komora_syncer.helpers.utils import build_tenant_name

import hmac
import hashlib
import time
import base64
import urllib.parse
import json
import time


import logging
# Log Warnings and higher errors from imported modules
logging.getLogger(requests.__name__).setLevel(logging.WARNING)
logging.getLogger('urllib3').setLevel(logging.WARNING)
logger = logging.getLogger(__name__)


class KomoraApi:
    def __init__(self):
        self.sites = []
        self.regions = []
        self.districts = []
        self.municipalities = []
        self.organizations = []
        self.contacts = []
        self.devices = []

    def checkout_url(model, params=""):
        private_key, sign_app = get_config()['komora']['PRIVATE_KEY'], get_config()[
            'komora']['SIGN_APP']
        komora_url = get_config()['komora']['KOMORA_API_URL']

        base_url = f"{komora_url}/{model}"
        signDate = str(int(time.time()))

        if params:
            params = f"?{params}&signApp={sign_app}&signDate={signDate}"
        else:
            params = f"?signApp={sign_app}&signDate={signDate}"

        digest = hmac.new(bytes(private_key, "utf-8"),
                          msg=str.encode(params), digestmod=hashlib.sha256).digest()
        signature = base64.b64encode(digest).decode()
        parsed_signature = urllib.parse.quote_plus(signature)

        result_url = f"{base_url}{params}&signHash={parsed_signature}"
        return result_url

    # TODO: catch errors
    def get_sites(self):
        if self.sites:
            return self.sites

        url = KomoraApi.checkout_url("Site", "")

        try:
            logger.info("Polling sites from Komora")
            response = requests.get(url)
            response.raise_for_status()
        except Exception as e:
            logger.error(f"Unable to poll sites from Komora\n{e}")
            logger.debug(f"URL: {url}")
            logger.debug(traceback.format_exc())
            raise e
        else:
            response = response.json()

        for site in response.get("data", []):
            self.sites.append(Site(**site))

        return self.sites

    def get_organizations(self):
        if self.organizations:
            return self.organizations

        filter = ''  # filter = 'filters.isCustomer=true'
        url = KomoraApi.checkout_url("Organization", filter)

        try:
            logger.info("Polling organizations from Komora")
            response = requests.get(url)
            response.raise_for_status()
        except Exception as e:
            logger.exception("Unable to poll organizations from Komora")
            logger.debug(f"URL: {url}")
            raise e
        else:
            response = response.json()

        # TODO: unique cids
        unique_cids_names = []
        for organization in response.get("data", []):
            cid_name = build_tenant_name(organization['clientId'], organization['name'])
            if cid_name not in unique_cids_names:
                unique_cids_names.append(cid_name)
            else:
                logger.critical(f"Duplicated CID and Name: {cid_name}")
                continue
            self.organizations.append(Organization(**organization))

        return self.organizations

    def get_regions(self):
        if self.regions:
            return self.regions

        api = 'Ruian/region'

        try:
            logger.info("Polling regions from Komora")
            for reg in KomoraApi.__get_paged_records(api):
                self.regions.append(Region(**reg))
        except Exception as e:
            logger.error(f"Unable to poll regions from Komora\n{e}")
            logger.debug(traceback.format_exc())
            raise e
        return self.regions

    def get_districts(self):
        if self.districts:
            return self.districts

        api = 'Ruian/district'

        try:
            logger.info("Polling districts from Komora")
            for dis in KomoraApi.__get_paged_records(api):
                self.districts.append(District(**dis))
        except Exception as e:
            logger.exception("Unable to poll districts from Komora")
            raise e

        return self.districts

    def get_municipalities(self):
        if self.municipalities:
            return self.municipalities

        api = "Ruian/municipality"

        try:
            logger.info("Polling municipalities from Komora")
            for muni in KomoraApi.__get_paged_records(api):
                self.municipalities.append(Municipality(**muni))
        except Exception as e:
            logger.exception("Unable to poll municipalities from Komora")
            raise e

        return self.municipalities

    def get_devices(self):
        if self.devices:
            return self.devices

        api = "Device"
        filter_params = 'Filters.IsActive=true'

        try:
            logger.info("Polling devices from Komora")
            for dev in KomoraApi.__get_paged_records(api, filter_params=filter_params):
                self.devices.append(Device(**dev))
        except Exception as e:
            logger.exception("Unable to poll devices from Komora")
            raise e

        return self.devices

    def post_devices(self, nb_devices):
        api = "ServiceRecord/SendDeviceData"
        url = KomoraApi.checkout_url(api)

        # nb_devices save to file
        # TODO: filepath - move to config
        timestr = time.strftime("%Y%m%d-%H%M%S")
        filename = f"{timestr}_devices.json"
        filepath = get_config()['common']['DEVICES_DATA_PATH']
        file = filepath+filename

        with open(file, 'w') as outfile:
            json.dump(nb_devices, outfile)

        try:
            logger.info("Posting devices to Komora")
            requests.post(url, files={'dataFile': open(file, 'rb')})
        except Exception as e:
            logger.exception("Unable to post devices to Komora")
            raise e

    def get_contacts(self):
        if self.contacts:
            return self.contacts

        filter = ''  # filter = 'filters.isCustomer=true'
        url = KomoraApi.checkout_url("Contact", filter)

        try:
            logger.info("Polling contacts from Komora")
            response = requests.get(url)
            response.raise_for_status()
        except Exception as e:
            logger.exception("Unable to poll contacts from Komora")
            logger.debug(f"URL: {url}")
            raise e
        else:
            response = response.json()

        for contact in response.get("data", []):
            self.contacts.append(Contact(**contact))

        return self.contacts

    def __get_page(api, page, page_size, filter_params=""):
        params = f"Page={page}&PageSize={page_size}"
        if filter_params:
            params = params + "&" + filter_params

        url = KomoraApi.checkout_url(api, params)

        try:
            response = requests.get(url)
            response.raise_for_status()
        except Exception as e:
            logger.exception("Unable to get page")
            logger.debug(f"URL {url}")
            raise e
        else:
            response = response.json()
            result = response.get('data', [])

            # Check number of all records
            total = response.get('total', 0)

            return result, int(total)

    def __get_paged_records(api, filter_params=""):
        page_size = 100
        page = 0

        # load first page
        record_list, total = KomoraApi.__get_page(
            api, page, page_size, filter_params=filter_params)

        # if less records are returned than total number of records -> get rest of records
        if len(record_list) < total:
            last_page = int((total/page_size))

            for page in range(1, last_page + 1):
                next_data = KomoraApi.__get_page(
                    api, page, page_size, filter_params=filter_params)[0]
                record_list.extend(next_data)

        if len(record_list) == total:
            return record_list
        # TODO: generate proper error
        else:
            raise Exception



from komora_syncer import __appname__

import os
import appdirs
import yaml

import configparser
import logging
logger = logging.getLogger(__name__)

os.environ["XDG_CONFIG_DIRS"] = "/etc"
CONFIG_DIRS = (
    appdirs.user_config_dir(__appname__),
    appdirs.site_config_dir(__appname__),
)
CONFIG_FILENAME = "config.yml"
LOGGING_FILENAME = "logging.ini"


def get_config():
    """
    Get config file and load it with yaml
    :returns: loaded config in yaml, as a dict object
    """
    if getattr(get_config, "cache", None):
        return get_config.cache

    if os.environ.get("CONFIG_FOLDER_PATH"):
        config_path = os.path.join(os.environ["CONFIG_FOLDER_PATH"], CONFIG_FILENAME)
    else:
        for d in CONFIG_DIRS:
            config_path = os.path.join(d, CONFIG_FILENAME)
            if os.path.isfile(config_path):
                break
    try:
        with open(config_path, "r") as config_file:
            conf = yaml.safe_load(config_file)
            get_config.cache = conf
            return conf
    except FileNotFoundError as e:
        logger.debug(e)
        if os.environ.get("CONFIG_FOLDER_PATH"):
            logger.error(
                "Configuration file {} not found.".format(
                    os.environ["CONFIG_FOLDER_PATH"]
                )
            )
        else:
            logger.error(
                "No configuration file can be found. Please create a "
                "config.yml in one of these directories:\n"
                "{}".format(", ".join(CONFIG_DIRS))
            )
        raise FileNotFoundError


def get_logging_path():
    """
    Get logging.ini file path
    :returns: path to logging.ini file
    """

    if os.environ.get("CONFIG_FOLDER_PATH"):
        log_conf_path = os.path.join(os.environ.get(
            "CONFIG_FOLDER_PATH"), LOGGING_FILENAME)
    else:
        for d in CONFIG_DIRS:
            log_conf_path = os.path.join(d, LOGGING_FILENAME)
            if os.path.isfile(log_conf_path):
                break

    try:
        with open(log_conf_path, "r") as log_file:
            return log_conf_path
    except FileNotFoundError as e:
        logger.debug(e)
        if os.environ.get("CONFIG_FOLDER_PATH"):
            logger.error(
                "Configuration file {} not found.".format(
                    log_conf_path
                )
            )
        else:
            logger.error(
                "No logging configuration file can be found. Please create a "
                "logging.ini in one of these directories:\n"
                "{}".format(", ".join(CONFIG_DIRS))
            )
        raise FileNotFoundError


class LoggingConfig():
    def __init__(self):
        super().__init__()
        self._config = LoggingConfig.load_config()

    def load_config():
        log_conf_file = get_logging_path()
        config = configparser.ConfigParser()

        try:
            if config.read(log_conf_file):
                pass
            else:
                raise FileNotFoundError
        except FileNotFoundError as e:
            logger.exception(f"Unable to open file {log_conf_file}")
            raise

        return config



import logging
from komora_syncer.connections.NetboxConnection import NetboxConnection


logger = logging.getLogger(__name__)


class NetboxCache():
    tenant_cache = {}
    device_cache = {}
    nb = NetboxConnection()
    nb.open()
    nb_client = nb.connection
    
    def __new__(cls):
        raise TypeError("This is a static class")
    
    @staticmethod
    def init_tenant_cache(force=False):
        if NetboxCache.tenant_cache and not force:
            return NetboxCache.tenant_cache
        
        try:
            tenants = NetboxCache.nb_client.tenancy.tenants.all()
            for tenant in tenants:
                if tenant.custom_fields.get("komora_id"):
                    NetboxCache.tenant_cache[tenant.custom_fields.get("komora_id")] = tenant   
        except Exception:
            logger.exception("Unable to initialize tenant cache")
            raise
    
    @staticmethod
    def init_device_cache(force=False):
        if NetboxCache.device_cache and not force:
            return NetboxCache.device_cache
        
        try:
            devices = NetboxCache.nb_client.dcim.devices.all()
            for device in devices:
                if device.custom_fields.get("komora_id"):
                    NetboxCache.device_cache[device.custom_fields.get("komora_id")] = device
        except Exception:
            logger.exception("Unable to initialize device cache")
            raise
                
    @staticmethod
    def get_tenant_cache():
        if not NetboxCache.tenant_cache:
            NetboxCache.init_tenant_cache()
        return NetboxCache.tenant_cache
    
    @staticmethod
    def get_device_cache():
        if not NetboxCache.device_cache:
            NetboxCache.init_device_cache()
        return NetboxCache.device_cache


class Address():
    def __init__(self, **kwargs):
        self.municipalityId = kwargs.get('municipalityId')
        self.municipalityName = kwargs.get('municipalityName')
        self.districtId = kwargs.get('districtId')
        self.districtName = kwargs.get('districtName')
        self.regionId = kwargs.get('regionId')
        self.regionName = kwargs.get('regionName')
        self.houseNumber = kwargs.get('houseNumber')
        self.referenceNumber = kwargs.get('referenceNumber')
        self.zip = kwargs.get('zip')
        self.streetName = kwargs.get('streetName')
        self.streetId = kwargs.get('streetId')
        self.text = kwargs.get('text')
        self.isActive = kwargs.get('isActive')
        self.id = kwargs.get('id')



from komora_syncer.config import get_config
import pynetbox
import logging
logger = logging.getLogger(__name__)


class NetboxConnection:
    def __init__(self):
        self.netbox_url = get_config()['netbox']['NETBOX_INSTANCE_URL']
        self.api_token = get_config()['netbox']['NETBOX_API_TOKEN']
        self.connection = None

    # https://pynetbox.readthedocs.io/en/latest/endpoint.html
    def open(self):
        try:
            # private_key_file='/path/to/private-key.pem',
            self.connection = pynetbox.api(
                self.netbox_url, token=self.api_token)
        except Exception as e:
            logger.exception("Could not establish connection to pynetbox")
            raise e



from komora_syncer.models.netbox.NetboxBase import NetboxBase
from slugify import slugify

import logging
logger = logging.getLogger(__name__)


class NbRegion(NetboxBase):
    # Types = region, district, municipality
    def __init__(self, komora_obj=None, type=None):
        # TODO:
        if type not in ['region', 'district', 'municipality']:
            raise Exception

        NetboxBase.__init__(self)
        self.api_object = None

        self.name = komora_obj.name
        self.slug = slugify(self.name)
        self.komora_id = komora_obj.id
        self.region_type = type

        # TODO: TODO: TODO: ZAHRANICI
        if self.slug == 'zahranicni-kraj':
            self.parent = None
            pass
        elif type == 'region':
            parent = "CZ"
            self.find_parent(parent, "state")
        elif type == 'district':
            parent = komora_obj.regionName
            self.find_parent(parent, "region")
        elif type == 'municipality':
            parent = komora_obj.districtName
            self.find_parent(parent, "district")

    def find_parent(self, parent, region_type):
        try:
            if parent:
                self.parent = self.netbox.connection.dcim.regions.get(
                    name=parent, cf_region_type=region_type, cf_komora_id__n=self.komora_id)

            else:
                self.parent = None
        except Exception as e:
            logger.exception(f"Parent region {parent} does not exists")
            self.parent = None

    def find_or_create(self):
        self.find()
        if not self.api_object():
            self.create()

        return self.api_object

    def find(self):
        if self.komora_id:
            try:
                netbox_region = self.netbox.connection.dcim.regions.get(
                    cf_komora_id=self.komora_id)

                if netbox_region:
                    self.api_object = netbox_region
                    return self.api_object
            except Exception as e:
                logger.exception(
                    f"Unable to find region by komora_id: {self.komora_id}")

    def create(self):
        try:
            netbox_region_param = self.get_params()
            netbox_region = self.netbox.connection.dcim.regions.create(
                netbox_region_param)

            logger.info("Region: %s created sucessfully", self.name)
            self.api_object = netbox_region
        except Exception as e:
            logger.exception("Unable to create netbox site: %s", self.name)
            raise e

        return self.api_object

    def update(self, nb_region):
        try:
            if nb_region.update(self.get_params()):
                self.api_object = nb_region
                logger.info(f"Region: {self.name} updated successfuly")
        except Exception as e:
            logger.exception(f"Unable to update region {self.name}")
            raise e

    def synchronize(self):
        region = self.find()

        if region:
            self.update(region)
        else:
            self.create()

    def get_params(self):
        params = {'name': self.name,
                  'slug': self.slug}

        if self.parent:
            params['parent'] = self.parent.id

        if self.komora_id:
            if type(params.get('custom_fields')) is dict:
                params['custom_fields']['komora_id'] = self.komora_id
            else:
                params['custom_fields'] = {"komora_id": self.komora_id}

        if self.region_type:
            if type(params.get('custom_fields')) is dict:
                params['custom_fields']['region_type'] = self.region_type
            else:
                params['custom_fields'] = {"region_type": self.region_type}
        return params



from komora_syncer.models.netbox.NetboxBase import NetboxBase
from slugify import slugify

import logging
logger = logging.getLogger(__name__)


class NbLocation(NetboxBase):
    def __init__(self, location_komora, parent_location_komora, nb_site):
        NetboxBase.__init__(self)
        self.name = str(location_komora.name or "").strip()
        self.slug = slugify(self.name)
        self.description = str(location_komora.description or "").strip()
        self.komora_id = location_komora.id
        self.api_object = None
        self.site = nb_site

        self.parent = None

        if parent_location_komora:
            nb_parent_location = NbLocation(
                parent_location_komora, None, nb_site)

            if nb_parent_location.find():
                self.parent = nb_parent_location.api_object
            else:
                logger.critical(
                    f"Parent object exists: {parent_location_komora.name}, but not foud in netbox")
                raise Exception(
                    f"Parent object exists: {parent_location_komora.name}, but not foud in netbox")

    def find_or_create(self):
        self.find()
        if not self.api_object():
            self.create()

        return self.api_object

    def find(self):
        try:
            netbox_location = self.netbox.connection.dcim.locations.get(
                name__ie=self.name, site_id=self.site.id)

            if netbox_location:
                self.api_object = netbox_location

        except Exception as e:
            logger.exception(
                f"Unable to get location {self.name} {self.site.name} from Netbox")
            raise(e)

        return self.api_object

    def create(self):
        try:
            params = self.get_params()
            netbox_location = self.netbox.connection.dcim.locations.create(
                params)

            if netbox_location:
                logger.info("location: %s created sucessfully", self.name)
                self.api_object = netbox_location
            else:
                logger.critical(
                    f"Unable to create locations {self.name}, site: {self.site_id}")
        except Exception as e:
            logger.exception("Unable to create netbox location: %s", self.name)
            raise e

        return self.api_object

    def update(self, nb_location):
        try:
            if nb_location.update(self.get_params()):
                self.api_object = nb_location
                logger.info(f"Location: {self.name} updated successfuly")
        except Exception as e:
            logger.exception(
                f"Unable to update location {self.name} in Netbox")
            raise e

    def synchronize(self):
        location = self.find()

        if location:
            self.update(location)
        else:
            self.create()

    def get_params(self):
        params = {
            'name': self.name,
            'slug': self.slug,
            "description": self.description,
            "site": self.site.id}

        if self.parent:
            params['parent'] = self.parent.id

        if self.komora_id:
            if type(params.get('custom_fields')) is dict:
                params['custom_fields']['komora_id'] = self.komora_id
            else:
                params['custom_fields'] = {"komora_id": self.komora_id}

        return params



"""Package configuration."""
from setuptools import find_packages, setup
from komora_syncer import __appname__

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

requirements = [
  'click==8.0.3',
  'pynetbox==7.0.1',
  'python-slugify==5.0.2',
  'appdirs==1.4.4',
  'pyyaml==6.0',
]

setup_requirements = [
    "pytest-runner",
]
test_requirements = [
    "pytest", "pytest-cov", "pytest-mock", "pynxos", "ipdb"
]

setup(
      name=__appname__,
      author="Jan Krupa",
      author_email="jan.krupa@cesnet.cz",
      classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Intended Audience :: Developers",
        "Development Status :: 4 - Beta",
        "Operating System :: Unix",
      ],
      description='Synchronize data between Komora and Netbox applications',
      long_description=long_description,
      long_description_content_type="text/markdown",
      install_requires=requirements,
      include_package_data=True,
      keywords="netbox,komora",
      packages=find_packages(),
      setup_requires=setup_requirements,
      test_suite="tests",
      tests_require=test_requirements,
      url="https://gitlab.cesnet.cz/701/done/netbox_komora_syncer",
      version='2.7',
      zip_safe=False,
      python_requires='>=3.6, <4',
      entry_points={
          'console_scripts': [
            'komora_syncer=komora_syncer.__main__:cli',
            ]
      },
     )
